/*
 * Copyright (c) 2013 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Fri Jan 18 2013
 */
#include "acrbulkit.h"

#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include <err.h>
#include <errno.h>
#include <libgen.h>
#include <limits.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>


/* --- usage --------------------------------------------------------------- */

/** Print usage on a channel.
 *
 * \param[in] channel   Output channel
 * \param[in] argv0     Program invocation name
 */

static void
usage(FILE *channel, char *argv0)
{
  fprintf(channel, "usage:  %s file program [args...]\n", basename(argv0));
}


/* --- noop_sighandler ----------------------------------------------------- */

static void
noop_sighandler(int sig)
{
}


/* --- unlock -------------------------------------------------------------- */

static void
unlock(const char *lock)
{
  while (unlink(lock)) {
    if (errno == EINTR) continue;
    warn("%s", lock);
    break;
  }
}


/* --- main ---------------------------------------------------------------- */

int
main(int argc, char *const argv[])
{
  extern char *optarg;
  extern int optind;
  char *argv0 = argv[0];

  struct sigaction act;
  sigset_t block, dflt;
  char host[PATH_MAX];
  char lock[PATH_MAX];
  struct timespec ts;
  struct timeval tv;
  siginfo_t info;
  struct stat sb;
  pid_t pid;
  int c, s;
  int gen;

  /* Parse command line options. __GLIBC__ will understand a leading '+' as
   * 'POSIXLY_CORRECT' and not try to interpret options beyond the first
   * non-option argument (here, the lock file). Other C libraries will
   * not interpret the '+' specially but it will not hurt either, since the
   * 'default' case will reject it. */
  while ((c = getopt(argc, argv, "+h")) != -1) {
    switch (c) {
      case 'h':
        usage(stdout, argv0);
        exit(0);

      case '?':
      default:
        usage(stderr, argv0);
        exit(2);
    }
  }
  argc -= optind;
  argv += optind;
  if (argc < 2) {
    usage(stderr, argv0);
    exit(2);
  }
  if (stat(argv[0], &sb)) err(2, "%s", argv[0]);

  /* enable SIGCHLD and block signals */
  act.sa_sigaction = NULL;
  act.sa_handler = noop_sighandler;
  sigemptyset(&act.sa_mask);
  act.sa_flags = SA_NOCLDSTOP|SA_RESTART;
  if (sigaction(SIGCHLD, &act, NULL))
    err(2, "sigaction");

  sigfillset(&block);
  if (sigprocmask(SIG_BLOCK, &block, &dflt)) err(2, "sigprocmask");


  /* get hostname and pid for unique lock file name */
  if (gethostname(host, sizeof(host))) host[0] = *"";
  pid = getpid();
  gettimeofday(&tv, NULL);
  srandom(tv.tv_usec);
  *lock = *"";

  /* lock */
  while(1) {
    /* generate a unique file name to defeat NFS attribute cache */
    gen = random();
    snprintf(lock, sizeof(lock), "%s:%d:%.64s:%d", argv[0], gen, host, pid);

    /* link and check st_nlink value */
    if (link(argv[0], lock)) err(2, "%s", lock);
    if (stat(lock, &sb)) err(2, "%s", lock);
    if (sb.st_nlink == 2) break;

    /* race lost: start over after some time */
    unlock(lock);

    ts.tv_sec = gen%10;
    ts.tv_nsec = 0;
    while (sigtimedwait(&block, &info, &ts) >= 0) {
      if (!info.si_signo) continue;
      exit(128 + info.si_signo);
    }
    if (errno != EAGAIN) err(2, "sigtimedwait");
  }

  /* spawn command */
  switch((pid = vfork())) {
    case -1:
      warn("fork");
      unlock(lock);
      exit(2);

    case 0:
      /* child */
      (void)setsid();
      (void)sigprocmask(SIG_SETMASK, &dflt, NULL);
      execvp(argv[1], argv+1);
      warn("execvp");
      _exit(127);
      break;
  }

  /* forward signals to child */
  do {
    if (sigwait(&block, &s)) {
      warn("sigwait");
      unlock(lock);
      exit(2);
    }
    if (s != SIGCHLD) kill(pid, s);
  } while(s != SIGCHLD);

  while (wait(&s) < 0) {
    if (errno == EINTR) continue;
    warn("wait");
    unlock(lock);
    exit(2);
  }

  /* unlock */
  unlock(lock);

  /* exit status */
  if (WIFEXITED(s)) return WEXITSTATUS(s);
  if (WIFSIGNALED(s)) return 128 + WTERMSIG(s);
  return 255;
}
